@@ -2,10 +2,11 @@ package Panda::Install;
 use strict;
 use warnings;
 use Config;
+use Cwd 'abs_path';
 use Exporter 'import';
 use Panda::Install::Payload;
 
-our $VERSION = '1.0.1';
+our $VERSION = '1.2.0';
 
 =head1 NAME
 
@@ -122,7 +123,11 @@ sub makemaker_args {
     delete $params{$_} for qw/SRC/;
     $params{OBJECT} = '$(O_FILES)' unless defined $params{OBJECT};
     
-    process_WIN32(\%params) if $win32;
+    if (my $shared_libs = $params{MODULE_INFO}{SHARED_LIBS} and $^O ne 'darwin') { # MacOSX doesn't allow for linking with bundles :(
+        my %seen;
+        @$shared_libs = grep {!$seen{$_}++} reverse @$shared_libs;
+        $params{LDFROM} .= ' '.join(' ', @$shared_libs) if @$shared_libs;
+    }
     
     delete $params{MODULE_INFO};
 
@@ -253,8 +258,19 @@ sub _apply_BIN_DEPS {
     $params->{PREREQ_PM}{$module}           ||= $installed_version;
     $params->{MODULE_INFO}{BIN_DEPS}{$module} = $installed_version;
     
-    _apply_BIN_DEPS_WIN32($params, $module) if $win32;
-
+    # add so/dll to linker list
+    my $shared_list = $params->{MODULE_INFO}{SHARED_LIBS} ||= [];
+    my $module_path = $module;
+    $module_path =~ s#::#/#g;
+    die "SHOULDN'T EVER HAPPEN" unless $module =~ /([^:]+)$/;
+    my $module_last_name = $1;
+    foreach my $dir (@INC) {
+        my $lib_path = "$dir/auto/$module_path/$module_last_name.$Config{dlext}";
+        next unless -f $lib_path;
+        push @$shared_list, abs_path($lib_path);
+        last;
+    }    
+    
     my $info = Panda::Install::Payload::module_info($module);
     
     if ($info->{INCLUDE}) {
@@ -305,23 +321,6 @@ sub _apply_BIN_DEPS {
     }
 }
 
-sub _apply_BIN_DEPS_WIN32 {
-    my ($params, $module) = @_;
-    
-    # add dll to linker list
-    my $list = $params->{MODULE_INFO}{WIN32}{DEPS_DLLS} ||= [];
-    my $module_path = $module;
-    $module_path =~ s#::#/#g;
-    die "SHOULDN'T EVER HAPPEN" unless $module =~ /([^:]+)$/;
-    my $last = $1;
-    foreach my $dir (@INC) {
-        my $dll_path = "$dir/auto/$module_path/$last.$Config{dlext}";
-        next unless -f $dll_path;
-        push @$list, $dll_path;
-        last;
-    }
-}
-
 sub process_BIN_SHARE {
     my $params = shift;
     my $bin_share = delete $params->{BIN_SHARE} or return;
@@ -358,13 +357,6 @@ sub process_BIN_SHARE {
     $pm->{$infopath} = '$(INST_ARCHLIB)/$(FULLEXT).x/info';
 }
 
-sub process_WIN32 {
-    my $params = shift;
-    my $deps_dlls = $params->{MODULE_INFO}{WIN32}{DEPS_DLLS};
-    return unless $deps_dlls and @$deps_dlls;
-    $params->{LDFROM} .= ' '.join(' ', @$deps_dlls);
-}
-
 sub attach_BIN_DEPENDENT {
     my $params = shift;
     my @deps = keys %{$params->{MODULE_INFO}{BIN_DEPS} || {}};
@@ -533,7 +525,7 @@ sub _string_merge {
         MY;
     use Config;
 
-    if ($^O eq 'MSWin32') {
+    if ($win32) {
         my $gcc_compliant = $Config{cc} =~ /\b(gcc|clang)\b/i ? 1 : 0;
         
         *dynamic_lib = sub {
@@ -621,7 +613,7 @@ options, ...etc of another. Panda::Install also lets you put source files in sub
 
 The params for Panda::Install are compatible with MakeMaker with some additions.
 
-Also it supports typemap inheritance and C-like XS synopsis.
+Also it greatly extends typemap functionality and adds C-like XS synopsis.
 
 =head1 SYNOPSIS
 
@@ -922,12 +914,12 @@ T_TYPE2 will have mycode1 inserted before mycode2 as if it was written
 
 =head3 Passing params
 
-You can pass params when inheriting typemaps. These params can be accessed in parent typemap via %p hash.
+You can pass params when inheriting typemaps. These params can be accessed in parent typemap.
 
     T_TYPE1
-        int $p{varname} = 150;
+        int $varname = 150;
         mycode1;
-        $p{expr};
+        $expr;
 
     T_TYPE2 : T_TYPE1(varname=myvar, expr="myvar = a + b")
         mycode2;
@@ -940,10 +932,10 @@ will result in (for input typemap)
         myvar = a + b;
         mycode2;
 
-=head2 TYPEMAP INIT CODE
+=head2 TYPEMAP INIT BLOCKS
 
-In OUTPUT typemaps you can use 'INIT: expr;' expressions. These expressions later will be moved to the top of the XS function like 
-INIT: section of XS function itself. It is useful for typemaps which want to predefine some variable, so that user has a chance
+In OUTPUT typemaps you can use C<INIT { block }> expressions. These blocks later will be moved to the top of the XS function like 
+C<INIT:> section of XS function itself. It is useful for typemaps which want to predefine some variable, so that user has a chance
 to change it. Such typemaps then use this variable in its code. For example:
 
     TYPEMAP
@@ -953,7 +945,10 @@ to change it. Such typemaps then use this variable in its code. For example:
     OUTPUT
     
     MY_TYPE
-        INIT: int lolo = 0;
+        INIT {
+            int lolo = 0;
+            // something else
+        }
         sv_setiv($arg, $var + lolo);
         
     
@@ -967,6 +962,21 @@ to change it. Such typemaps then use this variable in its code. For example:
     OUTPUT:
         RETVAL;
 
+C<INIT> blocks are not evaled by typemaps and goes through as-is, so that you cannot use C<$arg>, C<$var>, etc inside its code.
+
+=head2 TYPEMAP INIT_OUTPUT BLOCKS
+
+In OUTPUT typemaps you can use C<INIT_OUTPUT { block }> expressions. These blocks later will be moved to the top of the resulting
+output typemap code (after all inheritances). The are concatenated in reverse order, i.e. the most-base typemap's first.
+C<INIT_OUTPUT> is useful for example when you want to cancel all output-related typemap code and return something different.
+As an example, L<Panda::XS> backreferencing technique use it. These blocks are evaled normally.
+
+=head2 TYPEMAP DESTROY BLOCKS
+
+In INPUT typemaps you can use C<DESTROY { block }> expressions. These expressions are inserted into C<DESTROY> function, before
+any user-defined code in C<DESTROY> method is executed, and after input typemap runs, so that C<THIS> is visible.
+Blocks are inserted in order from derived to parent. These blocks are evaled normally.
+
 =head2 TYPEMAP CAST
 
 Sometimes the type of data you receive or return depends on something and therefore you cannot use certain input or output typemap.
