@@ -19,6 +19,13 @@ Panda::Install::ParseXS - Adds new features for TYPEMAPS and XS code
 our ($top_typemaps, $cur_typemaps);
 our %code_params;
 
+our $re_parens  = qr/(\((?:(?>[^()]+)|(?-1))*\))/;
+our $re_gtlt    = qr/(<(?:(?>[^<>]+)|(?-1))*>)/;
+our $re_braces  = qr/(\{(?:(?>[^{}]+)|(?-1))*\})/;
+our $re_quot1 = qr/"(?:[^"\\]+|\\.)*"/;
+our $re_quot2 = qr/'(?:[^'\\]+|\\.)*'/;
+our $re_quot  = qr/(?:$re_quot1|$re_quot2)/;
+
 sub map_postprocess {
     my $map = shift;
     my $is_output = $map->isa('ExtUtils::Typemaps::OutputMap') || 0;
@@ -28,29 +35,38 @@ sub map_postprocess {
     $code =~ s/\s+$//;
     $code =~ s/\t/    /g;
     #$code =~ s#^(.*)$#sprintf("%-80s%s", "$1", "/* $map->{xstype} */")#mge if $code;
-    my @attrs = qw/PREVENT_DEFAULT_DELETE_ON_EMPTY_DESTROY/;
-    my $initcode = '';
+    my @attrs = qw/PREVENT_DEFAULT_DESTROY/;
     
-    if ($map->{xstype} =~ s/^(.+?)\s+:\s+([^() ]+)\s*(\(((?:[^()]+|(?3))*)\))?$/$1/) {
+    if ($map->{xstype} =~ s/^(.+?)\s+:\s+([^() ]+)\s*(\(((?:$re_quot|[^()]|(?3))*)\))?$/$1/) {
         my $parent_xstype = $2;
         my $parent_params = $4;
         my $parent_map = $is_output ? outmap($parent_xstype) : inmap($parent_xstype);
         die "\e[31m No parent $parent_xstype found in $type map \e[0m" unless $parent_map;
         my $parent_code = $parent_map->code;
-        $initcode = $parent_map->{_init_code};
+        $map->{_init_code} = $parent_map->{_init_code};
         $map->{_attrs}{$_} = $parent_map->{_attrs}{$_} for @attrs;
         
         if ($parent_params and $parent_code) {
-            my @pairs = split /\s*,\s*/, $parent_params;
-            foreach my $pair (@pairs) {
-                my ($k,$v) = split /\s*=\s*/, $pair, 2;
-                $v //= '';
-                if (index($v, '"') == 0 and rindex($v, '"') == length($v) - 1) {
-                    substr($v, 0, 1, '');
-                    chop($v);
+            $parent_params .= ',';
+            $parent_params =~ s/^[\s,]+//;
+            my %p = (xstype => $parent_xstype);
+            while ($parent_params =~ s/^\s*((?:$re_quot|[^,])*)\s*,\s*//) {
+                my $pair = $1;
+                next unless defined $pair and $pair ne '' and $pair =~ /^((?:$re_quot|[^=])*)(?:=(.+))?$/;
+                my ($k,$v) = ($1, $2);
+                $v //= 1;
+                foreach my $quot ('"', "'") {
+                    for ($k, $v) {
+                        if (index($_, '"') == 0 and rindex($_, '"') == length($_) - 1) {
+                            substr($_, 0, 1, '');
+                            chop($_);
+                        }
+                    }
                 }
-                $parent_code = "    \${ \$p{'$k'} //= '$v'; \\''; }\n$parent_code";
+                $p{$k} = $v;
             }
+            my $pstr = parent_typemap_serialize_params(\%p);
+            $parent_code = "    \${\\\$self->eval_typemap($is_output, $pstr, 1)}";
         }
         
         if ($code =~ /TYPEMAP::SUPER\(\)\s*;/) {
@@ -66,34 +82,83 @@ sub map_postprocess {
         }
     }
     
-    # MOVE AWAY 'INIT: ...' code. It will be used in fetch_para to insert it into the top of the function
-    while ($code =~ s/^\s*INIT:(.+)//m) {
-        my $line = $1;
-        $line =~ s/^\s+//;
-        $line =~ s/\s+$//;
-        $initcode .= "$line\n";
-    }
-        
-    $map->{_init_code} = $initcode;
+    die "Panda::Install::ParseXS: the old-style 'INIT: ...' line found in a typemap code. If it's not your line, please consider updating Panda::XS"
+        if $code =~ /^\s*INIT:/m;
     
+    # MOVE AWAY 'INIT {...}' blocks. It will later be used in fetch_para to insert it into the top of the function
+    my $init_blocks = join("\n", extract_blocks('INIT', \$code));
+    $map->{_init_code} .= "$init_blocks\n" if $init_blocks;
+
     foreach my $attr (@attrs) {
         next unless $code =~ s/^\s*\@$attr\s*(?:=\s*(.+))?$//mg;
         $map->{_attrs}{$attr} = $1 // 1;
         $map->{_attrs}{$attr} =~ s/\s+$//;
     }
     
-    if ($is_output) {
-        # if code has '$arg = <something>' not in first line - prevent fuckin ExtUtils::ParseXS from adding '$arg = sv_newmortal()'
-        # triggered by $arg = NULL which must firstly be set by typemap itself. Move it on top if inheriting
-        #state $ppline = '    $arg = NULL; /* suppress xsubpp\'s pollution of $arg */';
-        my $found = ($code =~ s/^\s*\$arg\s*=\s*\NULL\s*;\s*$//gm); # remove previous guardians;
-        $code = "    \$arg = NULL;\n$code" if $found;
-        $code =~ s/\n\s*\n/\n/gm;
-    }
-    
     $map->code($code);
 }
 
+sub extract_blocks {
+    my ($blkname, $coderef) = @_;
+    return unless $$coderef;
+    return if index($$coderef, $blkname) == -1;
+    my @result;
+    while ($$coderef =~ s/^\s*$blkname\s*(?<CONTENT>$re_braces)//m) {
+        my $chunk = $+{CONTENT};
+        substr($chunk, 0, 1, '');
+        chop($chunk);
+        $chunk =~ s/^\s+//;
+        $chunk =~ s/\s+$//;
+        next unless $chunk;
+        push @result, $chunk;
+    }
+    return @result;
+}
+
+sub fix_newmortal {
+    my ($arg, $coderef) = @_;
+    # if code has '$arg = <something>' not in first line - prevent fuckin ExtUtils::ParseXS from adding '$arg = sv_newmortal()'
+    # triggered by $arg = NULL which must firstly be set by typemap itself. Move it on top if inheriting
+    #state $ppline = '    $arg = NULL; /* suppress xsubpp\'s pollution of $arg */';
+    my $found = ($$coderef =~ s/^\s*\Q$arg\E\s*=\s*\NULL\s*;\s*$//gm); # remove previous guardians;
+    $$coderef = "    $arg = NULL;\n$$coderef" if $found;
+}
+
+sub prettify {
+    my ($coderef) = @_;
+    return unless $$coderef;
+    $$coderef =~ s/\n\s*\n/\n/gm;
+    $$coderef .= "\n" unless substr($$coderef, -1, 1) eq "\n"; # typemap must end with \n
+}
+
+sub sync_type {
+    my ($type, $ntype, $subtype) = @_;
+    $type //= '';
+    return ($type, $type) unless $type;
+    unless ($subtype) {
+        $subtype = $type;
+        $subtype =~ s/\s+//g;
+        $subtype =~ s/\*$//;
+    }
+    unless ($ntype) {
+        $ntype = $type;
+        $ntype =~ s/\s+//g;
+        $ntype =~ s/\*/Ptr/g;
+    }
+    return ($ntype, $subtype);
+}
+
+sub parent_typemap_serialize_params {
+    my $hash = shift;
+    my $str = '{%$params,';
+    while (my ($k, $v) = each %$hash) {
+        s/#/\\#/g for $k, $v;
+        $str .= "qq#$k# => qq#$v#,";
+    }
+    chop($str);
+    return "$str, %p}"; # custom params to typemap are not rewritable by typemap's params to it's parent
+}
+
 sub outmap { return $cur_typemaps->get_outputmap(xstype => $_[0]) || $top_typemaps->get_outputmap(xstype => $_[0]); }
 sub inmap  { return $cur_typemaps->get_inputmap(xstype => $_[0]) || $top_typemaps->get_inputmap(xstype => $_[0]); }
 
@@ -103,7 +168,7 @@ use strict;
 use warnings;
 
     # pre process     # post process       # after preprocess
-my ($orig_fetch_para, $orig_print_section, $after_fetch_para);
+my ($orig_fetch_para, $orig_print_section, $after_fetch_para, $last_params);
 
 BEGIN {
     $orig_fetch_para = \&fetch_para;
@@ -167,7 +232,9 @@ sub fetch_para {
         
         $lines->[0] = $type;
         
+        $self->{_empty_func} = 0;
         if (!$lines->[1]) {{ # empty sub
+            $self->{_empty_func} = 1;
             my ($class, $func, $var);
             if ($sig =~ /^([^:]+)::([a-zA-Z0-9_\$]+)/) {
                 ($class, $func, $var) = ("$1*", $2, 'THIS');
@@ -175,7 +242,7 @@ sub fetch_para {
                 ($class, $func, $var) = ($2, $1, $3);
             } else { last }
             my $in_tmap = $self->{typemap}->get_inputmap(ctype => $class) or last;
-            if ($func eq 'DESTROY' and $var eq 'THIS' and $in_tmap->{_attrs}{PREVENT_DEFAULT_DELETE_ON_EMPTY_DESTROY}) {
+            if ($func eq 'DESTROY' and $var eq 'THIS' and $in_tmap->{_attrs}{PREVENT_DEFAULT_DESTROY}) {
                 splice(@$lines, 1, 0, ' ');
                 splice(@$linno, 1, 0, $linno->[0]);
             }
@@ -240,8 +307,8 @@ sub print_section {
     my $linno = $self->{line_no};
     
     # find typemap_in|outcast<>()
-    state $re_parens = qr/(\((?:(?>[^()]+)|(?-1))*\))/;
-    state $re_gtlt = qr/(<(?:(?>[^<>]+)|(?-1))*>)/;
+    my $re_parens = $Panda::Install::ParseXS::re_parens;
+    my $re_gtlt   = $Panda::Install::ParseXS::re_gtlt;
     my %gen_funcs;
     foreach my $row (['typemap_incast', 1], ['typemap_outcast', 0]) {
         my ($kword, $is_input) = @$row;
@@ -288,7 +355,7 @@ sub print_section {
                     my $tmcode = $tmap->code;
                     my $code = $self->$meth("qq\a$tmcode\a", $other);
                     my $arg_init = (!$is_input && $tmcode =~ /^\s*\$arg\s*=\s*\NULL\s*;\s*$/m) ? '' : ' = newSV(0)';
-                    my $tmfunc_code = _typemap_inline_func($tmfunc_name, $class, $code, $arg_init, $is_input, \@args, $tmap->{_init_code});
+                    my $tmfunc_code = $self->typemap_inline_func($tmfunc_name, $class, $code, $arg_init, $is_input, \@args, $tmap->{_init_code});
                     $gen_funcs{$tmfunc_name} = $tmfunc_code;
                 }
                 $expr =~ s/^\s*\(\s*/(aTHX_ /; # add interp to param for threaded perls
@@ -303,11 +370,11 @@ sub print_section {
     return $gen_code.$orig_print_section->($self, @_);
 }
 
-sub _typemap_inline_func {
-    my ($tmfunc_name, $class, $code, $arg_init, $is_input, $args, $tm_init) = @_;
+sub typemap_inline_func {
+    my ($self, $tmfunc_name, $class, $code, $arg_init, $is_input, $args, $tm_init) = @_;
     $code =~ s/^\s+//s;
     $code =~ s/\s+$//s;
-    my $additional_args = @$args ? ", ".join(", ", map { "$_->[0] $_->[1]" } @$args) : '';
+    my $additional_args = ($args && @$args) ? ", ".join(", ", map { "$_->[0] $_->[1]" } @$args) : '';
     
     if ($tm_init and @$args) {
         # if some custom typemap variable is defined and set in typemap INIT section, and is present in $args,
@@ -338,27 +405,119 @@ EOF
 }
 
 sub eval_output_typemap_code {
-    my ($self, $code, $other) = @_;
-    my ($Package, $ALIAS, $func_name, $Full_func_name, $pname) = @{$self}{qw(Package ALIAS func_name Full_func_name pname)};
-    my ($var, $type, $ntype, $subtype, $arg) = @{$other}{qw(var type ntype subtype arg)};
-    my %p;
-  
-    no warnings 'uninitialized';
-    my $rv = eval $code;
-    die "Error evaling typemap: $@\nTypemap code was:\n$code" if $@;
-    return $rv;
+    my ($self, $code, $args, $inner) = @_;
+    return $self->eval_typemap_code(1, $code, $args, $inner);
 }
 
 sub eval_input_typemap_code {
-    my ($self, $code, $other) = @_;
-    my ($Package, $ALIAS, $func_name, $Full_func_name, $pname) = @{$self}{qw(Package ALIAS func_name Full_func_name pname)};
-    my ($var, $type, $num, $init, $printed_name, $arg, $ntype, $argoff, $subtype) = @{$other}{qw(var type num init printed_name arg ntype argoff subtype)};
-    my %p;
+    my ($self, $code, $args, $inner) = @_;
+    return $self->eval_typemap_code(0, $code, $args, $inner);
+}
 
-    no warnings 'uninitialized';
-    my $rv = eval $code;
-    die "Error evaling typemap: $@\nTypemap code was:\n$code" if $@;
-    return $rv;
+sub eval_typemap_code {
+    my ($self, $_is_output_, $_code_, $params, $inner) = @_;
+    my %p = %{$params||{}};
+    my ($Package, $ALIAS, $func_name, $Full_func_name, $pname) = @{$self}{qw(Package ALIAS func_name Full_func_name pname)};
+    my ($var, $type, $num, $init, $printed_name, $arg, $argoff, $do_setmagic, $do_push) =
+        delete @p{qw(var type num init printed_name arg argoff do_setmagic do_push)};
+    my ($ntype, $subtype) = Panda::Install::ParseXS::sync_type($type, delete $p{ntype}, delete $p{subtype});
+    $last_params = $params if delete $p{_extract_params_};
+
+    my $rand = int(rand 2**63);
+    my $is_destroy = !$_is_output_ && $func_name eq 'DESTROY' and $var =~ /^THIS/;
+    my $is_empty_func = $self->{_empty_func};
+    my $has_threads = eval "use threads; 1;";
+    
+    my $_need_print_ = $_code_ =~ s/^\s*print\s+// ? 1 : 0;
+    $_code_ =~ s/["\a]\s*$//;
+    $_code_ =~ s/^\s*"//;
+    $_code_ =~ s/^qq\a//;
+    $_code_ = "qq\a$_code_\a";
+    $_code_ = "print $_code_" if $_need_print_;
+    
+    while (my ($k, $v) = each %p) { # place custom args in perl vars
+        $v //= '';
+        $_code_ = "my \$$k = '$v'; $_code_";
+    }
+    # need clean package everytime, because $code might set params which were not in %p, and therefore create a global var
+    # which would not be cleaned up
+    {
+        my $fn = $func_name || '';
+        my $pk = $Package || '';
+        my $st = $type || '';
+        s/[^a-zA-Z0-9_]/_/g for $fn, $pk, $st;
+        my $_pack = "TMEvalPack_${pk}_${fn}_${st}_".int(rand 2**63).'_'.length($_code_).($_is_output_ ? "O" : "I");
+        $_code_ = "{ package $_pack; $_code_ }";
+    }
+
+    my $_rv_;
+    {        
+        no strict;
+        no warnings 'uninitialized';
+        $_rv_ = eval $_code_;
+        die "Error evaling typemap: $@\nTypemap code was:\n$_code_" if $@;
+    }
+    
+    return $_rv_ if $inner; # don't post-process sub-typemaps separately, because INIT_OUTPUT, etc, behaviour won't be correct
+    return $_rv_ unless $arg; # skip primitive typemaps
+
+    if ($_is_output_) {
+        my $pre_blocks = join("\n", reverse Panda::Install::ParseXS::extract_blocks('INIT_OUTPUT', \$_rv_));
+        $_rv_ = "$pre_blocks\n$_rv_" if $pre_blocks;
+        Panda::Install::ParseXS::fix_newmortal($arg, \$_rv_);
+    } else {
+        my $destroy_blocks = join("\n", Panda::Install::ParseXS::extract_blocks('DESTROY', \$_rv_));
+        $_rv_ .= "\n$destroy_blocks" if $destroy_blocks and $is_destroy;
+    }
+    
+    Panda::Install::ParseXS::prettify(\$_rv_);
+    
+    return $_rv_;
+}
+
+sub eval_output_typemap {
+    my ($self, $args, $inner) = @_;
+    return $self->eval_typemap(1, $args, $inner);
+}
+
+sub eval_input_typemap {
+    my ($self, $args, $inner) = @_;
+    return $self->eval_typemap(0, $args, $inner);
+}
+
+sub eval_typemap {
+    my ($self, $is_output, $args, $inner) = @_;
+    my %params;
+    foreach my $k (qw/xstype ctype/) {
+        $params{$k} = delete $args->{$k} if exists $args->{$k};
+    }
+    my $method = $is_output ? 'get_outputmap' : 'get_inputmap';
+    my $typemap = $self->{typemap}->$method(%params);
+    my $dbgstr = $params{xstype} || $params{ctype};
+    die "\e[31m No typemap '$dbgstr' found for eval_typemap \e[0m" unless $typemap;
+    return $self->eval_typemap_code($is_output, '"'.$typemap->code.'"', $args, $inner);
+}
+
+sub output_typemap_param {
+    my ($self, $ctype, $pname) = @_;
+    return $self->typemap_param(1, $ctype, $pname);
+}
+
+sub input_typemap_param {
+    my ($self, $ctype, $pname) = @_;
+    return $self->typemap_param(0, $ctype, $pname);
+}
+
+sub typemap_param {
+    my ($self, $is_output, $ctype, $pname) = @_;
+    $self->eval_typemap($is_output, {
+        type  => $ctype,
+        ctype => $ctype,
+        arg   => 'dummy_arg',
+        var   => 'dummy_var',
+        _extract_params_ => 1,
+    }, 1);
+    return $last_params->{$pname};
 }
 
 package
@@ -404,7 +563,7 @@ sub new {
     my $self = $orig_onew->($proto, @_);
     Panda::Install::ParseXS::map_postprocess($self) unless ref $proto; # if $proto is object, it's cloning, no need to postprocess
     return $self;
-};
+}
 
 package
     ExtUtils::Typemaps::InputMap; # hide from pause
